<?php
/**
 * @file
 * Define DemandbaseCache class to override default drupal caching.
 */

// first load context module because we are using context_active_contexts() function which is defined in context.module file. This file define DemandbaseCache class which providing drupal caching and its called during drupal caching bootstrap step not after drupal bootstrap full.
// function demandbase_cache_boot(){
//   drupal_load('module', 'context');
// }
// drupal_load('module', 'context');
// //module_load_include('inc', 'context', 'context.core');

// //drupal_load('module', 'context.core');

// //drupal_load('module', 'context.plugins');
// if (!drupal_load('module', 'Demandbase cache')) {
//   include_once dirname(__FILE__) . '/demandbase_cache.module';
//   //dsm('demandbase_cache.module exits');
// }

/**
 * Defines a default cache implementation.
 *
 * This is Drupal's default cache implementation. It uses the database to store
 * cached data. Each cache bin corresponds to a database table by the same name.
 */
//variable_set('cache_class_cache_page', 'DemandbaseCache' );
class DemandbaseCache implements DrupalCacheInterface {
  protected $bin;

  /**
   * Constructs a new DrupalDatabaseCache object.
   */
  function __construct($bin) {
    $this->bin = $bin;    
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   */
  function getMultiple(&$cids) {
    try {
      // Garbage collection necessary when enforcing a minimum cache lifetime.
      $this->garbageCollection($this->bin);

      // When serving cached pages, the overhead of using db_select() was found
      // to add around 30% overhead to the request. Since $this->bin is a
      // variable, this means the call to db_query() here uses a concatenated
      // string. This is highly discouraged under any other circumstances, and
      // is used here only due to the performance overhead we would incur
      // otherwise. When serving an uncached page, the overhead of using
      // db_select() is a much smaller proportion of the request.
      
      //Demandbase cache module apply its hack here it just append the ip address of visitor to cid of page. 
      $cids[0] = $this->demandbase_cache_change_cid($cids[0]);
      //$cids[0] = self::demandbase_context_enabled($cids[0]);
      $result = db_query('SELECT cid, data, created, expire, serialized FROM {' . db_escape_table($this->bin) . '} WHERE cid IN (:cids)', array(':cids' => $cids));
      $cache = array();
      foreach ($result as $item) {
        $item = $this->prepareItem($item);
        if ($item) {
          $cache[$item->cid] = $item;
        }
      }
      $cids = array_diff($cids, array_keys($cache));
      return $cache;
    }
    catch (Exception $e) {
      // If the database is never going to be available, cache requests should
      // return FALSE in order to allow exception handling to occur.
      return array();
    }
  }

  /**
   * Garbage collection for get() and getMultiple().
   *
   * @param $bin
   *   The bin being requested.
   */
  protected function garbageCollection() {
    $cache_lifetime = variable_get('cache_lifetime', 0);

    // Clean-up the per-user cache expiration session data, so that the session
    // handler can properly clean-up the session data for anonymous users.
    if (isset($_SESSION['cache_expiration'])) {
      $expire = REQUEST_TIME - $cache_lifetime;
      foreach ($_SESSION['cache_expiration'] as $bin => $timestamp) {
        if ($timestamp < $expire) {
          unset($_SESSION['cache_expiration'][$bin]);
        }
      }
      if (!$_SESSION['cache_expiration']) {
        unset($_SESSION['cache_expiration']);
      }
    }

    // Garbage collection of temporary items is only necessary when enforcing
    // a minimum cache lifetime.
    if (!$cache_lifetime) {
      return;
    }
    // When cache lifetime is in force, avoid running garbage collection too
    // often since this will remove temporary cache items indiscriminately.
    $cache_flush = variable_get('cache_flush_' . $this->bin, 0);
    if ($cache_flush && ($cache_flush + $cache_lifetime <= REQUEST_TIME)) {
      // Reset the variable immediately to prevent a meltdown in heavy load situations.
      variable_set('cache_flush_' . $this->bin, 0);
      // Time to flush old cache data
      db_delete($this->bin)
        ->condition('expire', CACHE_PERMANENT, '<>')
        ->condition('expire', $cache_flush, '<=')
        ->execute();
    }
  }

  /**
   * Prepares a cached item.
   *
   * Checks that items are either permanent or did not expire, and unserializes
   * data as appropriate.
   *
   * @param $cache
   *   An item loaded from cache_get() or cache_get_multiple().
   *
   * @return
   *   The item with data unserialized as appropriate or FALSE if there is no
   *   valid item to load.
   */
  protected function prepareItem($cache) {
    global $user;

    if (!isset($cache->data)) {
      return FALSE;
    }
    // If the cached data is temporary and subject to a per-user minimum
    // lifetime, compare the cache entry timestamp with the user session
    // cache_expiration timestamp. If the cache entry is too old, ignore it.
    if ($cache->expire != CACHE_PERMANENT && variable_get('cache_lifetime', 0) && isset($_SESSION['cache_expiration'][$this->bin]) && $_SESSION['cache_expiration'][$this->bin] > $cache->created) {
      // Ignore cache data that is too old and thus not valid for this user.
      return FALSE;
    }

    // If the data is permanent or not subject to a minimum cache lifetime,
    // unserialize and return the cached data.
    if ($cache->serialized) {
      $cache->data = unserialize($cache->data);
    }

    return $cache;
  }

  /**
   * Implements DrupalCacheInterface::set().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    //demandbase cache module apply same hack here appends visitor ip address to cid of page.
    $cid = $this->demandbase_cache_change_cid($cid);
    //$cid = self::demandbase_context_enabled($cid);
    $fields = array(
      'serialized' => 0,
      'created' => REQUEST_TIME,
      'expire' => $expire,
    );
    if (!is_string($data)) {
      $fields['data'] = serialize($data);
      $fields['serialized'] = 1;
    }
    else {
      $fields['data'] = $data;
      $fields['serialized'] = 0;
    }

    try {
      db_merge($this->bin)
        ->key(array('cid' => $cid))
        ->fields($fields)
        ->execute();
    }
    catch (Exception $e) {
      // The database may not be available, so we'll ignore cache_set requests.
    }
  }

  /**
   * Implements DrupalCacheInterface::clear().
   */
  function clear($cid = NULL, $wildcard = FALSE) {
    global $user;

    if (empty($cid)) {
      if (variable_get('cache_lifetime', 0)) {
        // We store the time in the current user's session. We then simulate
        // that the cache was flushed for this user by not returning cached
        // data that was cached before the timestamp.
        $_SESSION['cache_expiration'][$this->bin] = REQUEST_TIME;

        $cache_flush = variable_get('cache_flush_' . $this->bin, 0);
        if ($cache_flush == 0) {
          // This is the first request to clear the cache, start a timer.
          variable_set('cache_flush_' . $this->bin, REQUEST_TIME);
        }
        elseif (REQUEST_TIME > ($cache_flush + variable_get('cache_lifetime', 0))) {
          // Clear the cache for everyone, cache_lifetime seconds have
          // passed since the first request to clear the cache.
          db_delete($this->bin)
            ->condition('expire', CACHE_PERMANENT, '<>')
            ->condition('expire', REQUEST_TIME, '<')
            ->execute();
          variable_set('cache_flush_' . $this->bin, 0);
        }
      }
      else {
        // No minimum cache lifetime, flush all temporary cache entries now.
        db_delete($this->bin)
          ->condition('expire', CACHE_PERMANENT, '<>')
          ->condition('expire', REQUEST_TIME, '<')
          ->execute();
      }
    }
    else {
      if ($wildcard) {
        if ($cid == '*') {
          db_truncate($this->bin)->execute();
        }
        else {
          db_delete($this->bin)
            ->condition('cid', db_like($cid) . '%', 'LIKE')
            ->execute();
        }
      }
      elseif (is_array($cid)) {
        // Delete in chunks when a large array is passed.
        do {
          db_delete($this->bin)
            ->condition('cid', array_splice($cid, 0, 1000), 'IN')
            ->execute();
        }
        while (count($cid));
      }
      else {
        db_delete($this->bin)
          ->condition('cid', $cid)
          ->execute();
      }
    }
  }

  /**
   * Implements DrupalCacheInterface::isEmpty().
   */
  function isEmpty() {
    $this->garbageCollection();
    $query = db_select($this->bin);
    $query->addExpression('1');
    $result = $query->range(0, 1)
      ->execute()
      ->fetchField();
    return empty($result);
  }

  /**
   *
   * change cache id (cid) for anonymous user based on their ip.
   * @param string type $key
   */
  public function demandbase_cache_change_cid($key = NULL) {
    $enable_deamandbase_cache_anonymouse = variable_get('demandbase_activate_cache_for_anonymous', '0');
    if ($enable_deamandbase_cache_anonymouse != 2) {
      return $key;
    }
    $enable_demandbase_cache_paths = variable_get('demandbase_activate_cache_path_list', '0');    
    if (!$enable_demandbase_cache_paths ) {
      $key = $key . '+' . ip_address();
      return $key;
    }
    $enable_demandbase_cache_list = variable_get('demandbase_cache_path_list', '');
    if (empty($enable_demandbase_cache_list)) {
      return $key;
    }      
    $paths = explode("\n", $enable_demandbase_cache_list);    
    $path_matched = $this->validate_path($paths);
    if ($path_matched) {
       $key = $key . '+' . ip_address();
    }
      return $key;
  }

  function validate_path($pages) {
    $path = $this->drupal_get_path_alias($_GET['q']);
    // Compare with the internal and path alias (if any).
    $page_match = $this->match($path, $pages, TRUE);
    if ($path != $_GET['q']) {
      $page_match = $page_match || $this->match($_GET['q'], $pages, TRUE);
    }
    return $page_match;
  }

  function drupal_get_path_alias($path) {
    //@TODO
    return $path;
  }

  function match($subject, $patterns, $path = FALSE) {
    static $regexps;
    $match = FALSE;
    $positives = $negatives = 0;
    $subject = !is_array($subject) ? array($subject) : $subject;
    foreach ($patterns as $pattern) {
      if (strpos($pattern, '~') === 0) {
        $negate = TRUE;
        $negatives++;
      }
      else {
        $negate = FALSE;
        $positives++;
      }
      $pattern = ltrim($pattern, '~');
      if (!isset($regexps[$pattern])) {
        if ($path) {
          $regexps[$pattern] = '/^(' . preg_replace(array('/(\r\n?|\n)/', '/\\\\\*/', '/(^|\|)\\\\<front\\\\>($|\|)/'), array('|', '.*', '\1' . preg_quote(variable_get('site_frontpage', 'node'), '/') . '\2'), preg_quote($pattern, '/')) . ')$/';
        }
        else {
          $regexps[$pattern] = '/^(' . preg_replace(array('/(\r\n?|\n)/', '/\\\\\*/'), array('|', '.*'), preg_quote($pattern, '/')) . ')$/';
        }
      }
      foreach ($subject as $value) {
        if (preg_match($regexps[$pattern], $value)) {
          if ($negate) {
            return FALSE;
          }
          $match = TRUE;
        }
      }
    }
    // If there are **only** negative conditions and we've gotten this far none
    // we actually have a match.
    if ($positives === 0 && $negatives) {
      return TRUE;
    }
    return $match;
  } 

  // public static function demandbase_context_enabled($cid){
  //   $demandbase_conditions = &drupal_static(__FUNCTION__);
  //   if(empty($demandbase_conditions)){
  //     $result = variable_get('context_status',''); //this variable return disabled contexts
  //     $disabled_contexts = preg_grep("/1/", $result);
  //     $enabled_contexts = db_select('context')
  //                           ->fields('context', array('conditions'))
  //                           ->condition('name', array_keys($disabled_contexts), 'NOT IN')
  //                           ->execute();
  //     //$demandbase_context = array();
  //     while ($enabled_context =  $enabled_contexts->fetchField()) {
  //       $demandbase_context[] = self::preg_grep_keys("/^demandbase_/", unserialize($enabled_context));
  //     }
  //     //flaten conditions
  //     foreach ($demandbase_context as  $value) {
  //       foreach ($value as $v) {
  //         $demandbase_conditions[] = $v;  
  //       }
         
  //     }
  //     if (!empty($demandbase_conditions)) {
  //       $cid = $cid . '+' .ip_address();
  //       return $cid;
  //     }
  //     return $cid;
  //   }else{
  //     if (!empty($demandbase_conditions)) {
  //       $cid = $cid . '+' .ip_address();
  //       return $cid;
  //     }
  //     return $cid;
  //   }
  // }
  // public static function preg_grep_keys($pattern, $input, $flags = 0) {
  //   $keys = preg_grep($pattern, array_keys($input), $flags);
  //   return $keys;
  // }
}